;; Generated package description from smalltalk-mode.el  -*- no-byte-compile: t -*-
(define-package "smalltalk-mode" "4.0.0.20210113.195524" "Major mode for the GNU Smalltalk programming language" 'nil :url "https://elpa.gnu.org/packages/smalltalk-mode.html" :authors '(("Steve Byrne")) :maintainer '("Derek Zhou" . "derek@3qin.us"))
